import java.lang.*;
import java.io.*;
import java.util.*;

public class AdvancedDownload extends Download implements Runnable
{
	private SegmentedFile destination;
    private long size;
	private Date time;
	
    String fileType;

	private DownloadThread downloadThread[];
	private int	connections;
	private int segments;
	
	private boolean	hasDone;	
		
	////////////////////////////////////////////////////////////////////////////////
	
    public AdvancedDownload( String location, String target, Hashtable options )
    throws DownloadException
    {
        this.location    = location;
        this.target      = target;
        this.options     = options;

		try {
	        if( options.containsKey("FILE_TYPE") ) {
		    	fileType = (String)options.get("FILE_TYPE");
			} else {
				fileType = "BINARY";
			}	

	        if( options.containsKey("CONNECTIONS") ) {
	        	connections = ((Integer)options.get("CONNECTIONS")).intValue();
				//	TODO:	check value
			} else {
				connections = 1;
			}
			
			if( options.containsKey("SEGMENTS") ) {
	        	segments = ((Integer)options.get("SEGMENTS")).intValue();
				//	TODO:	check value
			} else {
				segments = 1;
			} 

   		    File targetFile = new File( target );
   		    if( targetFile.length() > 0 ) {
		        if( !targetFile.canWrite() ) {
                    throw new DownloadException( "Download was aborted because can't write to " + target.toString() + ", check file attributes and retry." );
   		        }
   		        resume = true;
   		        size = targetFile.length(); 
   		        time = new Date( targetFile.lastModified() );
   		    } else {
   		        resume = false;
   		        size = 0;
   		        time = null;
   		    }
			
			hasDone = false;
			
		} catch( NullPointerException e ) {
		    throw new DownloadException( "Unrecoverable Error" );
		}
	
	}

    public void start()
    throws DownloadException
    {
    	if( !isIdle() ) {
    		return;
    	}
    	else {
	        hasDone = false;

            try {
            	Connection probe = Connection.connectTo( location, options );

                if( probe == null ) {
                    throw new DownloadException("Unable to connect to server.");
    		    }
    		    
    		    if( resume )
    		    try {
                    if( !probe.checkSize( size ) ) {
                        throw new DownloadException("Download has been aborted since source and destination sizes do not match, if you still want to proceed you must delete the local file.");
                    }

                	if( !probe.checkTime( time ) ) {
                        throw new DownloadException("Download has been aborted since source and destination dates do not match, if you still want to proceed you must delete the local file.");
                    }

                    //  destination = new SegmentedFile( target, size, segments );
                    
                    throw new DownloadException("Resuming of advanced downloads is not implemented yet.");
                    
                } catch( ConnectionException e ) {
                    throw e;
                } 
                else 
                try {
                    destination = new SegmentedFile( target, size, segments );
                
                } catch( IOException e ) {
                	throw new DownloadException("Unable to open target file");
                } 
            
                downloadThread = new DownloadThread[connections];

    			try {
    			        
                	downloadThread[0] = new DownloadThread( probe, destination, fileType );
    				downloadThread[0].start();

    			} catch( ConnectionException e ) {
    				//	Se nemmeno la prima connessione riesce ad iniziare il trasferimento non ha senso
    				//	tentare con le successive (questo e' insito nella semantica delle connection)
    				//	pertanto il download deve abortire.

    				probe.close();
    				throw new DownloadException("Unable to start download.");
    			}

    			for( int c=1; c<connections; c++ )
    			try {
    				downloadThread[c] = new DownloadThread( probe.clonate(), destination, fileType );
    				downloadThread[c].start();

    			} catch( ConnectionException e ) {
    				//	Questa eccezione puo essere stata causata dal tentativo di clonare
    				//	la connessione o dal tentativo di iniziare il download
    				//	in entrambi i casi si assume che non sia possibile generare ulteriori
    				//	threads per cui si lascia che continuino quelli gia'
    				//	avviati senza generarne altri.
    			
    				downloadThread[c] = null;

    				connections = c;
    			}

    			//	TODO:	fare partire un thread che a intervalli di tempo abbastanza lunghi
    			//			tenti di aggiungere connessioni fino a soddisfare il numero richiesto
    			//			dall'utente.
                //          Al momento il thread parte ma il metodo run() non fa niente.
                
                (new Thread( this ) ).start();
                
                return;

            } catch( NullPointerException e ) {
                throw new DownloadException("Unrecoverable error.");
            } catch( ConnectionException e ) {
            	//	TODO:	Exception description should be translated
                throw e;
            }
        }
    }
    
    private boolean	isIdle() {
    
		if( downloadThread == null ) {
		    return true;
		} 
		else {
		    boolean retVal = downloadThread[0] == null ? true : downloadThread[0].isIdle();
		    
		    for( int c=1; c<connections; c++ ) 
        	    retVal = retVal && downloadThread[c] == null ? true : downloadThread[c].isIdle();
            
            return retVal;
        }
    }		
	
    public void run() {
        //  Actually no monitoring is done so this thread can quit here.

        return;
    }
    
    public void stop()
    throws DownloadException
    {
		for( int c=0; c<connections; c++ ) {
			downloadThread[c].abort();
		}
		
		hasDone = downloadThread[0].hasDone();
        for( int c=1; c<connections; c++ ) 
        	hasDone = hasDone && downloadThread[c].hasDone();
        
        return;
    }

}