import java.lang.*;
import java.util.*;

abstract public class Connection extends java.lang.Object
{
    protected boolean   connected;
    protected Hashtable options;

    public Connection() {
	}

    static public Connection connectTo( String location, Hashtable options )
    throws ConnectionException
    {
        try {
            LocationParser locationParser = new LocationParser( location );

            String protocol = locationParser.protocol().toUpperCase();

            if( protocol.equals("FTP") )
            try {
                String  host = locationParser.host();
                int     port = locationParser.port();
                String  user = locationParser.username();
                String  pass = locationParser.password();
                String  file = locationParser.path();

                if( user == null ) {
                    user = "ANONYMOUS";
                    
                    if( options.containsKey( "EMAIL_ADDR" ) )
                        pass = (String)options.get( "EMAIL_ADDR" );
                    else
                        pass = "";
                }
                
                if( options.containsKey( "FILE_TYPE" ) ) {
                    if( ((String)options.get( "FILE_TYPE" )).equals("TEXT" ) ) {
                        FTPConnection newConnection = new FTPConnection( host, port, user, pass, file, options );
                        newConnection.setParameters( "AN", "F", "S" );
                        return newConnection;
                    }
                    if( ((String)options.get( "FILE_TYPE" )).equals("BINARY" ) ) {
                        FTPConnection newConnection = new FTPConnection( host, port, user, pass, file, options );
                        newConnection.setParameters( "I", "F", "S" );
                        return newConnection;
                    }
                } else { //...defaults to binary.
                    FTPConnection newConnection = new FTPConnection( host, port, user, pass, file, options );
                    newConnection.setParameters( "I", "F", "S" );
                    return newConnection;
                }
            }
            catch( ConnectionException e ) {
                throw e;
            }
            
            throw new ConnectionException( "Unknown protocol:" + protocol );

        } catch( LocationParserException e ) {
            throw new ConnectionException( "Parser exception" );
        }

    }

    abstract public Connection clonate() throws ConnectionException;

    abstract public void close();

    public boolean isConnected() {
		return connected;
    }

    abstract public boolean canResume();
    abstract public boolean canGetSize();
    abstract public boolean canGetTime();
    
    abstract public long getSize() throws ConnectionException;
    abstract public Date getTime() throws ConnectionException;
 
    abstract public boolean checkSize( long size ) throws ConnectionException;
    abstract public boolean checkTime( Date date ) throws ConnectionException;
    
    abstract public boolean skip( long numBytes ) throws ConnectionException;
    
    abstract public boolean download( long numBytes ) throws ConnectionException;
    abstract public boolean download() throws ConnectionException;
    
    abstract public boolean upload( long numBytes ) throws ConnectionException;
	abstract public boolean upload() throws ConnectionException;
	
    abstract public int read() throws ConnectionException;
    
    abstract public void write( int byteValue ) throws ConnectionException;

}