import java.lang.*;

public class DownloadThread extends java.lang.Thread
{
    public SegmentedFile segmentedFile;
    
	public Connection connection;
    
    private Segment segment;
    
    private boolean isIdle;     //  Indica lo stato del thread
	private boolean hasDone;	//	Indicazione che il thread ha completato il suo compito
    private boolean mustQuit;	//	Indicazione di terminazione forzata esternamente dal metodo abort()
    
	private int fileType;
	private final static int BINARY = 0x0000;
	private final static int TEXT = 0x0010;
	
	Thread  thread;
	
    public DownloadThread( Connection connection, SegmentedFile segmentedFile, String type ) 
    throws DownloadException
    {
    	this.connection = connection;
    	this.segmentedFile = segmentedFile;
    	
    	if( type.equals( "BINARY" ) ) fileType = BINARY;
    	else 
    	if( type.equals( "TEXT" ) ) fileType = TEXT;
    	else 
    	throw new DownloadException( "Unknown file type" );
    	
    	segment = null;
    	hasDone = false;
        isIdle  = true;
        
		//{{INIT_CONTROLS
		//}}
	}
    
    public void run()
    {
	   	do	{

    		if( ( segment = segmentedFile.firstFreeSegment() ) == null ) {
				hasDone = true;
	            isIdle = true;
	            return;
			} 
			else {
			    isIdle = false;

    			try {
    			    connection.skip( segment.getBase() + segment.getOffset() );
    			    connection.download( segment.getSize() - segment.getOffset() );
    			
    			} catch( ConnectionException e ) {
    			    //  todo:   download of this segment cannot start
    			    return;			
    			}
			
    			switch( fileType )
    			{
    			case TEXT:
    				hasDone = false;
    				receiveTextSegment();
    				break;
    			case BINARY:
    				hasDone = false;
    				receiveBinarySegment();
    				break;
    			default:
    				hasDone = false;
    				receiveBinarySegment();
    				break;
    			}
		    }
		
		} while( !hasDone && !mustQuit );
    }
    	
	private void receiveTextSegment() {
        try {
          	int nextByte;
            int otherByte;
            
            while( !hasDone )
            {
            	if( mustQuit ) {
            		return;
            	}
            
                switch( nextByte = connection.read() ) 
                {
                case (int)-1:		
                    segment.flush();
                	hasDone = true;
                	break;
                case (int)'\r':
                	if( ( otherByte = connection.read() ) == '\f' ) {
                    	segment.write( '\n' );
                    } else {
						segment.write( nextByte );
						segment.write( otherByte );
                    }
					break;
				default:
					segment.write( nextByte );
					break;
                }
           	}
 		
       	} catch( ConnectionException e ) {      
            return;
        } catch( SegmentException e ) {      
            return;
        }

    }
   
	private void receiveBinarySegment() {
        try {
        	int nextByte;
            
            while( !hasDone )
            {
            	if( mustQuit ) {
            		return;
            	}
            
                nextByte = connection.read();

                if( nextByte == -1 ) {
                    segment.flush();
                    hasDone = true;
                }
                else {
                    segment.write( nextByte );
                }
            }
 		
        } catch( ConnectionException e ) {      
            return;
        } catch( SegmentException e ) {      
            return;
        }
    }

    public boolean isIdle() {
 		return isIdle;
    }

    public boolean hasDone() {
 		return hasDone;
    }

    public boolean abort() {
 			mustQuit = true;
 			
 		return !hasDone;
    }

		//{{DECLARE_CONTROLS
		//}}
}