//  FTPConnection
//
//  last rev.   8/jan/2000  17:41

import java.lang.*;
import java.util.*;
import java.net.*;

public class FTPConnection extends Connection implements Cloneable
{
    private String  host;
    private int     port;
    private String  user;
    private String  pass;
    private String  file;

    private String  type;
    private String  stru;
    private String  mode;

    private UserPI  userPI;         //  Incapsula gli streams di controllo e implementa il protocollo
    private UserDTP userDTP;        //  Incapsula gli streams dati

    public FTPConnection( String host, int port, String user, String pass, String file, Hashtable options )
    throws ConnectionException
    {
        this.options = options;

        this.host = host;
        this.port = port;
        this.user = user;
        this.pass = pass;
        this.file = file;
        
        connect();
        
        type = "I";     //  Default to Image type
	    stru = "F";     //  Default to File structure
	    mode = "S";     //  Default to Stream mode
	}

    public void setParameters( String type, String stru, String mode )
    {
        //  TODO:   Add check...
    
        this.type = type;
	    this.stru = stru;
	    this.mode = mode;
	}
    
    private void connect() 
    throws ConnectionException
    {
        try {
            userPI = new UserPI( options );
        
            userPI.connect( host, port, user, pass );
    	    
    	} catch( ConnectionException e ) {
    	    
	        throw e;
	    
	    }
	}
    
	public Connection clonate()
    throws ConnectionException
    {
        try {
            FTPConnection newConnection = (FTPConnection)this.clone();
        
            newConnection.connect();
            
            return newConnection;
        } 
        catch( CloneNotSupportedException e ) {
            throw new ConnectionException( "Unable to clonate connection, this error should not appear, contact the author." );
        }
    }

	public void close()
    {
        try {
            userPI.abort();
            userDTP.close();
            userPI.disconnect();
        } catch( ConnectionException e ) {
            //  TODO:   Exception should be handled more properly :)
            return;
        }
    }

    public boolean canResume() {
        return userPI != null ? userPI.canResume() : false;
    }
    
    public boolean canGetSize() {
        return userPI != null ? userPI.canGetFileSize() : false;
    }
    
    public boolean canGetTime() {
        return userPI != null ? userPI.canGetFileTime(): false;
    }
    
    public long getSize() 
    throws ConnectionException
    {
        return userPI != null ? userPI.getFileSize(file) : -1L;
    }
    
    public Date getTime() 
    throws ConnectionException
    {
        return userPI != null ? userPI.getFileTime(file) : null;
    }
    
    public boolean checkSize( long size ) 
    throws ConnectionException
    {
        return userPI != null ? userPI.getFileSize(file) == size : false;
    }
    
    public boolean checkTime( Date date ) 
    throws ConnectionException
    {
        return userPI != null ? (userPI.getFileTime(file)).equals(date) : false;
    }
    
    private long skipBytes;
        
    public boolean skip( long numBytes ) 
    throws ConnectionException
    {
        if( userPI != null ) {
            skipBytes = numBytes;
    	    return userPI.canResume();
    	} else {
    	    return false;
    	}
    }

    public boolean download() 
    throws ConnectionException
    {
    	return ( userDTP = userPI.download( file, type, stru, mode, skipBytes ) ) != null;
    }

    public boolean download( long numBytes ) 
    throws ConnectionException
    {
    	return ( userDTP = userPI.download( file, type, stru, mode, skipBytes, numBytes ) ) != null;
    }

    public boolean upload() 
    throws ConnectionException
    {
        //  todo:   not implemented yet

    	return false;
    }

    public boolean upload( long numBytes ) 
    throws ConnectionException
    {
        //  todo:   not implemented yet

    	return false;
    }

    public void write( int byteValue ) 
    throws ConnectionException
    {
        //  todo:   not implemented yet
    }

    public int read() 
    throws ConnectionException
    {
        try {
            return userDTP.read();
        } 
        catch( ConnectionException e ) {
            throw e;
        }
    }

}