import java.lang.*;
import java.io.*;

public class Segment extends java.lang.Object
{
	private SegmentedFile owner;
	
	private long	start;
	private long 	end;
	private long 	pointer;	//	The pointer indicates the absolute address of the first 
	                            //  empty byte in the segment when the segment is empty 
	                            //  pointer == start while when the segment is full pointer == end

	private int     bufferSize;
    private byte    buffer[];
	private int     bufferIndex;
	
	public Segment( SegmentedFile owner, long start, long length, long pointer, int bufferSize ) {
    	this.owner = owner;
		
		this.start	= start;
		this.end	= start + length;
		
		if( pointer >= start && pointer <= end )
			this.pointer = pointer;
		else
			this.pointer = start;

		this.locked = false;
		
		this.bufferSize = bufferSize;
		this.buffer = new byte[ bufferSize ];
	    this.bufferIndex = 0;
	}
    
	public Segment( SegmentedFile owner, long start, long length, long pointer ) {
	    this(owner, start, length, start, 8 * 1024 );
    }

	public Segment( SegmentedFile owner, long start, long length ) {

		this(owner, start, length, start, 8 * 1024 );
	}

    public int read() 
	throws SegmentException
	{
		try	{
			owner.lock();
			
			owner.seek(pointer);
			
			int	value = owner.readByte();
			
			owner.unlock();
		
			return value;
			} 
		catch( IOException e ) {
			throw new SegmentException();
		}
	}

	public void write( int value ) 
	throws SegmentException
	{
	    byte byteValue = (byte)value;
	
	    if( bufferIndex < bufferSize ) {
	        buffer[ bufferIndex++ ] = byteValue;
		} else 
		try {
			owner.lock();
			
			owner.seek(pointer);
			
			owner.write(buffer);
						
			owner.unlock();
			
			pointer += bufferSize;
			} 
		catch( IOException e ) {
			throw new SegmentException();
		}
	}

    public void flush() 
	throws SegmentException
	{
	    try {
			owner.lock();
			
			owner.seek(pointer);
			
			for( int i=0; i<bufferIndex; i++ )
			    owner.writeByte(buffer[i]);
						
			owner.unlock();
			
			pointer += bufferIndex;
			} 
		catch( IOException e ) {
			throw new SegmentException();
		}
	}

    public long getBase() {
        return start;
    }

    public long getOffset() {
        return pointer - start;
    }

    public long getSize() {
        return end - start;
    }
    
    ///////////////////////////////////////////////////////////////////////////////
		
	private boolean locked;
	
	public synchronized boolean lock() {
		
		if( locked == false ) {
			locked = true;
			return true;
		} else {
			return false;
		}
	}
	
	public synchronized boolean unlock() {
		
		if( locked == true ) {
			locked = false;
			return true;
		} else {
			return false;
		}
	}
	
}