import java.lang.*;
import java.util.*;
import java.io.*;

public class SegmentedFile extends java.io.RandomAccessFile
{
	private int		segments;
	private Segment	segment[];
	
	private long size;
	private Date time;
	
	public SegmentedFile( String path, long size, int segments ) 
	throws IOException
	{
		super( path, "rw" );
		
		try {
		    File file = new File( path );
		    
		    if( file.exists() ) {
    		    size = file.length();
    		    time = new Date( file.lastModified() );
		    } else {
    		    size = 0;
    		    time = null;
		    }
		} catch( NullPointerException e ) { // If the pathname argument is null
		    size = 0;
		    time = null;
		}
		
		this.segments = segments;               
		this.segment = new Segment[segments];

		int segmentsSize = (int)Math.ceil( (float)size / (float)segments );

		for( int s=0; s<segments-1; s++ ) 
		{
			segment[s] = new Segment( this, s * segmentsSize, segmentsSize );
		}
			segment[segments-1] = new Segment( this, (segments-1) * segmentsSize, size );
		
		this.locked = false;
	}
	
	////////////////////////////////////////////////////////////////////////////////

    public long getSize() {
        return size;
    }

    public Date getTime() {
        return time;
    }

	public synchronized Segment firstFreeSegment() 
	{
		for( int i=0; i<segments; i++ ) 
			if( segment[i].lock() ) {
				return segment[i];
			}
	
		return null;
	}

	public synchronized Segment lastFreeSegment() 
	{
		for( int i=segments-1; i>=0; i-- ) 
			if( segment[i].lock() ) {
				return segment[i];
			}
	
		return null;
	}

	public synchronized Segment nextFreeSegment( int current ) 
	{
		for( int i=current; i<segments; i++ ) 
			if( segment[i].lock() ) {
				return segment[i];
			}
	
		return null;
	}

	////////////////////////////////////////////////////////////////////////////////
	
	private boolean	locked = true;
	
	public synchronized boolean lock() 
	{
		try {
			while( locked != false ) 
				this.wait();
				
			locked = true;
			return true;

		} catch( IllegalMonitorStateException e ) {
			return false;
		} catch( InterruptedException e ) {
			return false;
		}	
		
	}
	
	public synchronized boolean unlock() 
	{
		if( locked == true ) {
			locked = false;
	
			try {
				this.notify();
				return true;
			} catch( IllegalMonitorStateException e ) {
				return false;
			}

		} else {
			return false;
		}
	}

}
