//  SimpleDownload
//
//  last rev.   30/12/1999  20:09

import java.lang.*;
import java.io.*;
import java.util.*;

public class SimpleDownload extends Download implements Runnable
{
	private Connection connection;
    private RandomAccessFile destination;

	private long size;              //  Numero di bytes gia ricevuti == lungezza logica del file locale
	private Date time;

	private boolean translate;      //  I files non binari richiedono la traduzione di alcuni caratteri
	                                //  per cui tra le opzioni deve essere fornita l'inicazione del tipo

	private String type;            //  Al momento puo' essere solo TEXT che indica un file di testo ascii

	////////////////////////////////////////////////////////////////////////////////
	
	private boolean	connected;	    //  Questo flag indica se e' stata fatta la connessione ma non dice nulla
                                    //  se questa sia ancora attiva o meno. La caduta si rileva solo dalle eccezioni.

	private Thread  thread;         //  Ogni download e'un thread indipendente dall'applicazione principale
                                    //  l'applicazione mantiene un riferimento al download mentre questo e'
                                    //  un riferimento al suo thread in esecuzione.
	private boolean isIdle;     	//  Indica lo stato del thread
	private boolean hasDone;		//  Indica se il thread ha finito il suo compito
					        		//	Settando a true questo flag si forza la terminazione.
							
	////////////////////////////////////////////////////////////////////////////////
	
    public SimpleDownload( String location, String target, Hashtable options )
    throws DownloadException
    {
	    this.location = location;
        this.target = target;
        this.options = options;

	    try {
	        File targetFile = new File( target );
	        
	        size = targetFile.length();  			            //  Il controllo sull'esistenza non serve
	                                                            //  perche' in quel caso length ritorna 0L
		    if( size > 0 ) {

                time = new Date( targetFile.lastModified() );

                if( !options.containsKey("RESUME") ) {			//  Se non c'e' un'indicazione di come procedere il
		                                                        //  download deve abortire senza modificare il file.
		            throw new DownloadException( "Target file already exists and is not empty, download was aborted since no resume indication found in options." );
		        }
		        else try {
                    resume = ((Boolean)options.get("RESUME")).booleanValue();
		        } catch( NullPointerException e ) {
		            throw new DownloadException( "Target file already exists and is not empty, download was aborted since resume indication found in options was invalid." );
		        }

		        if( !targetFile.canWrite() ) {
                    throw new DownloadException( "Download was aborted because can't write to " + target.toString() + ", check file attributes and retry." );
                }

		    }
		    else {
		        resume = false;
		    }

		} catch( NullPointerException e ) {
		    throw new DownloadException( "Download was aborted because target file name is invalid." );
		}

        if( options.containsKey("FILE_TYPE") ) {
            translate = true;
            type = (String)options.get("FILE_TYPE");
        }
        else {
            translate = false;
        }

        thread = null;
        isIdle = true;
        hasDone = false;
	}

    public void start()
    throws DownloadException
    {
    	if( !isIdle ) {	
    		return;
    	}
    	else
        try {

            connection = Connection.connectTo( location, options );

            if( connected = ( connection != null ) ) {

                try {
                    if( size!=0 && !connection.checkSize( size ) ) {
                        throw new DownloadException("Download has been aborted since source and destination sizes do not match, if you still want to proceed you must delete the local file.");
                    }

                } catch( ConnectionException e ) {
                    //	TODO:	to be handled properly
                    throw e;
                }

                try {
                    if( time!=null && !connection.checkTime( time ) ) {
                        throw new DownloadException("Download has been aborted since source and destination dates do not match, if you still want to proceed you must delete the local file.");
                    }

                } catch( ConnectionException e ) {
                    //	TODO:	to be handled properly
                    throw e;
                }
   		
   				if( resume ) 
            		if( !connection.skip( size ) ) {
                		throw new DownloadException("Download was aborted because server can't resume transfer.");
            		}
                
                thread = new Thread(this);
                hasDone = false;
            	isIdle = false;
                
                connection.download();
                
                thread.start();
            }
            else {
                throw new DownloadException("Download was aborted because there is no connection to the server.");
            }

        } catch( ConnectionException e ) {
        
        	//	TODO:	Exception description should be translated if necessary
            
            throw new DownloadException(e.description);

        } catch( NullPointerException e ) {
            
            throw new DownloadException("Download was aborted because of an unrecoverable error.");
        }
    }
	
    public void stop()
    throws DownloadException
    {
        hasDone = true;
        return;
    }

	////////////////////////////////////////////////////////////////////////////////
    
    public void run()
    {
		if( translate ) {
		    if( type.equals("TEXT") ) receiveTextFile();
    		else {
    			receiveBinaryFile();
    		}
			//	TODO:	Add other translations here...
		}
		else {
			receiveBinaryFile();
		}
	}
	
	public void receiveTextFile() {
	    try {
            Writer writer = new FileWriter( target, resume );

            try {
                
                while( !hasDone )
                {
                    int nextChar = connection.read();

                    if( nextChar == -1 ) {
                        hasDone = true;
                    }
                    else {
                        writer.write( nextChar );
                    }
                }

            } catch( ConnectionException e ) {
                //  TODO:   Maybe I should write a better handler :D
            } catch( IOException e ) {
                //  TODO:   Maybe I should write a better handler :D
            }

            writer.close();

            connection.close();

            isIdle = true;

            return;

        } catch( FileNotFoundException e ) {    //  TODO:   questa eccezione viene scatenata se non e'possibile aprire il file destinazione.
            return;
        } catch( IOException e ) {              //  If a generic I/O error occurs.
            return;
        }
	}
   
	public void receiveBinaryFile() {
        try {
            OutputStream outputStream = new FileOutputStream( target, resume );

            try {
                
                while( !hasDone )
                {
                    int nextByte = connection.read();

                    if( nextByte == -1 ) {
                        hasDone = true;
                    }
                    else {
                        outputStream.write( nextByte );
                    }
                }

            } catch( ConnectionException e ) {
                //  TODO:   Maybe I should write a better handler :D
            } catch( IOException e ) {
                //  TODO:   Maybe I should write a better handler :D
            }

            outputStream.close();

            connection.close();

            isIdle = true;

            return;

        } catch( FileNotFoundException e ) {    //  TODO:   questa eccezione viene scatenata se non e'possibile aprire il file destinazione.
            return;
        } catch( IOException e ) {              // if an I/O error occurs.
            return;
        }

    }

}