//	UserDTP
//
//	last rev: 6-gen-2000    10:50

import java.lang.*;
import java.net.*;
import java.io.*;

public class UserDTP extends java.lang.Object implements Runnable
{
    private ServerSocket srvSocket;
    private Socket dtpSocket;

    private InputStream dcIn;
    private OutputStream dcOut;
    
    private boolean connected;			//  Indica che e' stata fatta l'accept e che si sta traferendo il file
    private boolean download;        	//  Rappresenta la direzione del trasferiemnto, true indica 
    								    //	dal server remoto al file system locale 
    								    //  questo flag non ha significato se connected e' false.
    private long transferSize;
    private long transferred;

    public UserDTP( int srvSoTimeout )
    throws ConnectionException
    {
        try {
            srvSocket   = new ServerSocket( 0 );	//  Creates a socket and binds it to any available port

            srvSocket.setSoTimeout( srvSoTimeout );

            dcIn = null;
            dcOut = null;

            this.transferSize = -1;         //  Trasfer Until server closes connection
            this.transferred = 0;           //  Bytes transferred so far...

            connected   = false;

        } catch( SocketException e ) {

        	throw new ConnectionException("Unknown Socket Exception.");			//  SoTimeout Error

        } catch( IOException e ) {                     	 						//  if an I/O error occurs
        																		//	when opening the socket.
            throw new ConnectionException("Unable to Open Data Connection.");
        }

		//{{INIT_CONTROLS
		//}}
	}
    
    
    public void upload( long transferSize ) 
    throws ConnectionException
    {
        accept( transferSize, false );
    }
    
    public void upload() 
    throws ConnectionException
    {
        accept( -1L, false );
    }
    
    public void download( long transferSize ) 
    throws ConnectionException
    {
        accept( transferSize, true );
    }
    
    public void download() 
    throws ConnectionException
    {
        accept( -1L, true );
    }
    
    public void accept( long transferSize, boolean download ) 
    {
        this.transferSize = transferSize;       //  Total number of bytes to be transferred from the start position
        this.transferred = 0;					//  Bytes transferred so far

        this.download = download;
            
        (new Thread(this)).start();

        //  Le eccezioni scatenate da questo metodo DEVONO fare abortire il trasferimento
    }
    
    public void run() {
        try {
            dtpSocket = srvSocket.accept();
            
            dcIn  = dtpSocket.getInputStream();
            dcOut = dtpSocket.getOutputStream();
        
            connected = true;						
        
        } catch( InterruptedIOException e ) {   
            // throw new ConnectionException("Unable to open data connection, accept timeout.");   //  Accept Timeout
        } catch( IOException e ) {                      
            // throw new ConnectionException("Unknown data connection exception.");                //  I/O Error
        }
    }
    
    public String getLocalHost() {
        
        InetAddress addr;
        
        try {
            addr = InetAddress.getLocalHost();
        } catch( UnknownHostException e ) {
            addr = null;
        }
        
        if( addr != null ) {
            return addr.getHostAddress();
        } else {
            return null;
        }
    }

    public int getLocalPort() {
    
        return srvSocket.getLocalPort();
    }

    public boolean hasDone() {
        return transferSize == -1 ? false : transferSize - transferred <= 0;
    }

    public int read() 
    throws ConnectionException
    {
        int retVal;

        try {
            if( hasDone() ) {           //  Check to see if this read extends past the specified number of bytes
                retVal = -1;            //  in this case we must signal an eof to the caller
            } else {
                retVal = dcIn.read();    
            
                if( retVal != -1 ) {
                    transferred++;
                }
            }
            
            return retVal;

        } catch( NullPointerException e ) {
            throw new ConnectionException("Null pointer exception, data connection unavailable.");
        } catch( java.io.IOException e ) {
            throw new ConnectionException("Interrupted Data Connection.");
        }
    }

    public void write( int byteValue ) 
    throws ConnectionException
    {
        //  ACTUALLY DOES NOTHING !!!
        /*
        if( download )
        	throw new ConnectionException("Can't Write to Data Connection");
        
        try {
            if( !connected )
                throw new ConnectionException("Data Connection Unavailable");
        
            if( hasDone() ) {                                                   //  Check to see if this read extends past the specified number of bytes
                throw new ConnectionException("Interrupted Data Connection");   //  TODO:   must be handled more gracefully
            } else {
                dcOut.write( byteValue );

                transferred++;     	//  Se il bytes non e' stato traferito 
                					//	non si raggiunge mai questo statement
            }
        
        } catch( java.io.IOException e ) {
        
            throw new ConnectionException("Interrupted Data Connection");
        }
        */
    }

    public void close() 
    throws ConnectionException
    {
        try {
            dcIn.close();
            dcOut.close();
        
            connected = false;
        
        } catch( java.io.IOException e ) {
        
            throw new ConnectionException("Unknown Data Connection Exception");
        }
    }
		//{{DECLARE_CONTROLS
		//}}
}