//  UserPI
//  
//  Last Rev:	2/1/2000

import java.lang.*;
import java.io.*;
import java.net.*;
import java.util.*;

public class UserPI extends java.lang.Object implements Runnable
{
    private Terminal terminal;  //  A terminal used to echo the commands sent to the server ad it's replies
    private boolean echo;       //  If this is false terminal echo is unavailable

    private Hashtable options;  //  A Reference to an Hashtable containing some connection options
    private int soTimeout;      //  Timeout in millisecondi
    private int soRetries;      //  Tentativi prima di scatenare una ConnectionException per fallimento
    
    private InetAddress host;   //  Server address
    private int port;           //  Server port
    private String user;        //  Username
    private String pass;        //  Password
    
    private InputStream ccIn;        //  Control Connection Input for reading messages sent from server PI
    private OutputStream ccOut;       //  Control Connection Output for writing messages to the server PI
    
    private Vector lastReply;   //  Il comando REPLY() riceve tutte le linee relative all'ultimo messaggio
                                //  di risposta e le mette in questo vettore che fa da buffer per
                                //  quei comandi che successivamente alla REPLY vogliono esaminarne il contenuto.
                                
    private boolean canREST;    //  Server can resume ?             ( this value is useless if connected is false )
    private boolean canSIZE;    //  Server can return files size ?  ( this value is useless if connected is false )
    private boolean canMDTM;    //  Server can return files time ?  ( this value is useless if connected is false )
    
    private UserDTP userDTP;    //  La connessione dati non e' necessario che sia attiva per tutto il tempo
    
	//  Nei flag che seguiranno e' codificato lo stato degli oggetti della classe UserPI.
    //  Un oggetto rappresenta una connessione con un server, il suo ciclo di vita parte dalla
    //  creazione mediante il costruttore che viene eseguito completamente offline
    //  al termine ( eventuali problemi terminano con un'eccezione ) il nuovo userPI
    //  e' in stato connected == false ( gli altri flag sono irrilevanti ) e solo dopo che viene
    //  eseguito il metodo connect, e solo se va a buon fine, si passa nello stato connected == true.
    //  ( Nota: connected == true non dice nulla sullo stato della connessione che potrebbe cadere in
    //  qualunque momento, cosa di cui ci si accorge solo mediante le eccezioni scatenate dalle READ/WRITE )
    //  Nello stato connected == true, si ha idle == true fintanto che non viene iniziato 
    //  un download o un upload mediante i metodi omonimi, da allora e fino a quando non viene ricevuto 
    //  il messaggio di completamento ( sia con successo che con insuccesso ) l'userPI si trova nello stato 
    //  connected ==true, idle == false che indica che c'e' un download in corso e che quindi 
    //  non ne possono cominciare di nuovi.
    //  completed riflette l'esito dell'ultimo trasferimento effettuato, il suo valore ha senso solo se idle == true.
    
    private boolean connected;      //  Indica se si e' stati connessi o disconnessi dal ServerPI
    private boolean idle;           //  Indica se c'e' un trasferimento attivo
    private boolean completed;      //  Indica se l'ultimo trasferimento ha completato con successo.
                                
    public UserPI( Hashtable options ) 
    {
        this.options = options;

        if( options.containsKey( "TERMINAL" ) ) 
        {
            terminal = (Terminal)options.get( "TERMINAL" );
        }
        
        if( terminal != null ) {
            echo = true;
        } else {
            echo = false;
        }
    
        if( options.containsKey( "CONNECTION_TIMEOUT" ) ) 
        try {
            soTimeout = 1000 * ( (Integer)options.get( "CONNECTION_TIMEOUT" ) ).intValue();
        } catch( NumberFormatException e ) {
            soTimeout = 60000;
        } 
    
        if( options.containsKey( "CONNECTION_RETRIES" ) ) 
        try {
            soRetries = ( (Integer)options.get( "CONNECTION_RETRIES" ) ).intValue();
        } catch( NumberFormatException e ) {
            soRetries = 10;
        } 
    
        this.canREST = false; 
        this.canSIZE = false;
        this.canMDTM = false;
    
        userDTP = null;
        
        this.connected = false;
        this.idle = true;
        this.completed = true;
	
		//{{INIT_CONTROLS
		//}}
	}
    
    public boolean echoOn() {
		if( terminal != null ) echo = true;
		return echo;		
    }

    public boolean echoOff() {
		if( terminal != null ) echo = false;
		return echo;		
    }

    public void connect( String host, int port, String username, String password )
    throws ConnectionException 
    {
        Socket  piSocket = null;	    // Questa inizializzazione e' richiesa dal compilatore symantec
                                        // perche ?...boh !!! 
		if(!connected)
        try {
            idle = false;
            
            this.host = InetAddress.getByName( host );
            this.port = port;
            this.user = username;
            this.pass = password;
        	
        	int retries = 0;
        	if( soRetries < 1 ) {
                throw new ConnectionException("The number of retries specified is incorrect.");
            }
        
            while( !connected )
            if( retries++ < soRetries )
	            try {
	                connected = (piSocket = new Socket( this.host, this.port )) != null;
	        	} catch( IOException e ) {  
	    	        try {   
	    	            Thread.sleep(soTimeout);
	                } catch( InterruptedException ie ) {
	                	continue;
	                }
	            } 
            else {
                throw new ConnectionException("Protocol Interpreter was unable to open control connection.");
            }
	
			try {
				piSocket.setSoTimeout( soTimeout );
            } catch( SocketException e ) {
                throw new ConnectionException("Unknown socket timeout exception.");
            }
            
            // InputStream iStream = piSocket.getInputStream();                    
            // ccIn  = new InputStreamReader( iStream );               //  Control Connection Input Stream Reader
            
            ccIn = piSocket.getInputStream();
            
            // OutputStream oStream = piSocket.getOutputStream();                  
            // ccOut = new OutputStreamWriter( oStream );              //  Control Connection Output Stream Writer
            
            ccOut = piSocket.getOutputStream();
            
            if( !REPLY("220") ) {
                throw new ConnectionException( "Server refused connection." );
	        }
            
            if( !USER( user ) ) {
                throw new ConnectionException("Login failed, bad username.");
	        }
            
            if( !PASS( pass ) ) {
                throw new ConnectionException("Login failed, bad password.");
	        }
	        
	        if( !FEAT() ){                  //  Server doesn't implements FEAT or has no ext.features
                canSIZE = false;
                canMDTM = false;
                canREST = false;
            } 
            else                            //  Determining ext. featrures supported
            try {
                Enumeration e = lastReply.elements();
                
                while( e.hasMoreElements() )
		        {
		            String replyLine = ((String) e.nextElement()).toUpperCase();
                    
                    if( replyLine.startsWith(" ") ) 
                    {
                        if( replyLine.startsWith( " REST" ) )		//  Expected line:  " REST STREAM"
                            canREST = replyLine.substring( 6, 12 ).equals("STREAM");
                        
                        if( replyLine.startsWith( " SIZE" ) )       //  Expected line:  " SIZE"
                            canSIZE = true;
                                    
                        if( replyLine.startsWith( " MDTM" ) )       //  Expected line:  " MDTM"
                            canMDTM = true;
                    }
                }
            
            } catch( NullPointerException e ) {
                //  TODO:   Something went wrong with interpretation of features
                throw new ConnectionException("Protocol Interpreter was unable to determine server features, null pointer exception.");
            } catch( NoSuchElementException e ) {
                //  Actually there is no reason to reach this point
                //  since nextElement is never called without before calling hasMoreElements
            }
            
            if( !SYST() ) {                                             
                throw new ConnectionException( "Unknown system on host." );
        	}
        	
            completed   = true;     //  Anche se non ho trasferito niente non ho nemmeno avuto insuccessi
                                    //  questa assunzione e' importatne perche alcuni siti lasciano connettersi
                                    //  ma non scaricare... ehm...questa idea la devo approfondire meglio :D

        	connected   = true;     //  Now this Client Protocol Interpreter is connected to his Server
        	
        	idle        = true;     //  But sill NO download in progress
    
        } catch( UnknownHostException e ) {     
            //  InetAddress.getByName() - if no IP address for the host could be found
            throw new ConnectionException("Unknown Host, maybe address is incorrect.");
	    
	    } catch( IOException e ) {              
	        //  Socket.getInputStream()
            //  Socket.getOutputStream()
            throw new ConnectionException("Socket I/O exception.");
	    }
		else {
			// TODO:	Handle the case in which the connect method is invoked again 
            
            // actually does nothing !            
		}
    
    }
    
    public UserDTP download( String path, String type, String stru, String mode, long start, long length ) 
    throws ConnectionException
    {
        if( !idle ) 
            throw new ConnectionException("A transfer is already in progress...");
        else
        try {
            idle = false;
            
            userDTP = new UserDTP( soTimeout );
            
            if( !PORT( userDTP.getLocalHost(), userDTP.getLocalPort() ) ) {                                                           
                throw new ConnectionException("PORT command failed.");
            } 

            if( !TYPE( type ) ) {                  
                throw new ConnectionException("TYPE command failed.");
    	    } 
            if( !STRU( stru ) ) {
                throw new ConnectionException("STRU command failed.");
    	    } 
            if( !MODE( mode ) ) {
                throw new ConnectionException("MODE command failed.");
    	    } 

            if( start > 0L )
            if( !REST( start ) ) {  
	        	throw new ConnectionException("REST command failed.");
	        } 
    	        	    
            userDTP.download( length );
    
			if( RETR( path ) ) 
            {
                (new Thread(this)).start();
                
                return userDTP;
            } 
            else {
                throw new ConnectionException("Server was unable to start download.");
    	    }
        
        } catch( ConnectionException e ) {
            //  TODO:   If underlying exceptions must be intercepted it must be done here otherwise just rethrow them
            throw   e;                       
        }
    }
    
    public UserDTP download( String path, String type, String stru, String mode, long start ) 
    throws ConnectionException
    {
        return download( path, type, stru, mode, start, -1L ); 
    }
    
    public UserDTP download( String path, String type, String stru, String mode ) 
    throws ConnectionException
    {
        return download( path, type, stru, mode, 0L, -1L ); 
    }
    
    public void run()
    {
        try {
            completed = REPLY( "250" );     //  Questo thread aspetta la terminazione del download nel metodo privato REPLY
                                            //  Se viene invocato il metodo pubblico Abort questo fa si che il server invii 
                                            //  un messaggio di terminazione del download che sblocca questo thread
            userDTP.close();
            
        } catch( ConnectionException e ) {
            completed =false;
        }
        
        idle = true;
    }
    
    public boolean abort() 
    throws ConnectionException
    {
        return ABOR();          //  ABOR non ha valore di ritorno in quanto la sua reply e' il messaggio
                                //  di completamento del download che viene gestito dal thread in attesa
                                //  return e' sempre true, se ci sono dei problemi nell'invio del comando 
                                //  viene scatenata un'eccezione.
    }
    
    public boolean canResume() {
        return connected ? canREST : false;
    }
    
    public boolean canGetFileSize() {
        return connected ? canSIZE : false;
    }
    
    public boolean canGetFileTime() {
        return connected ? canMDTM : false;
    }
    
    public long getFileSize( String path ) 
    throws ConnectionException
    {
        long retVal = -1;
        
        try {
        
            if( SIZE( path ) ) {
                String replyLine = ((String) lastReply.firstElement()).toUpperCase();
                long retval = Long.parseLong( replyLine.substring(4) );
            }
            else retVal = -1;  //  unable to determine file size
        
        } catch( NullPointerException e ) {
            retVal = -1;
        } catch( NoSuchElementException e ) {
            retVal = -1;
        } 
        
        return retVal;
    }
    
    public Date getFileTime( String path ) 
    throws ConnectionException
    {
        Date    retVal = null;
        
        try {
        
            if( SIZE( path ) ) {
                
                String  replyLine = ((String) lastReply.firstElement()).toUpperCase();
                
                int year    = Integer.parseInt(replyLine.substring(4,8));
                int month   = Integer.parseInt(replyLine.substring(8,10));
                int day     = Integer.parseInt(replyLine.substring(10,12));
                int hours   = Integer.parseInt(replyLine.substring(12,14));
                int minutes = Integer.parseInt(replyLine.substring(14,16));
                int seconds = Integer.parseInt(replyLine.substring(16,18));
                
                Calendar    fileTime = Calendar.getInstance();
                            fileTime.set( year, month, day, hours, minutes, seconds );
                            
                retVal = fileTime.getTime();
            }
            else retVal = null;  //  unable to determine file modification time
        
        } catch( NullPointerException e ) {
            retVal = null;
        } catch( NoSuchElementException e ) {
            retVal = null;
        }
        
        return retVal;
    }
    
    public boolean disconnect()			//	Per ora dovrebbe funzionare ma va risista la semantica
    throws ConnectionException
    {
        try {
            if( !connected ) 
                return false;
            
            if( !idle ) {
            	abort();
        	}
        	
            ccIn.close();
            ccOut.close();
            
            connected = false;
            
            return !connected;
            
        } catch( java.io.IOException e ) {
            
            throw new ConnectionException("I/O Error.");
        }
    }
    
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    private void WRITE( String message )							//	Writes a String on the ccOut
    throws ConnectionException
    {
        if( message != null ) 
        try {
            ccOut.write( message.getBytes(), 0, message.length() );
            
            if(echo) terminal.write( message );

        } catch( java.io.InterruptedIOException e ) {
            throw new ConnectionException("Control connection interrupted.");
        } catch( java.io.IOException e ) {
            throw new ConnectionException("Unknown control connection exception.");
        } 
    }
    
    private void WRITELN( String line )								//	...same as above but appends <CR><LF>
    throws ConnectionException
    {
        if( line != null ) 
        try {
            ccOut.write( (line.concat( "\r\n" )).getBytes(), 0, line.length() + 2 );
            
            if(echo) terminal.writeln( line );

        } catch( java.io.InterruptedIOException e ) {
            throw new ConnectionException("Control connection interrupted.");
        } catch( java.io.IOException e ) {
            throw new ConnectionException("Unknown control connection exception.");
        } 
    }
    
    private String READLN() 										//	Read a line from ccIn...doesn't return
    throws ConnectionException										//	  until reaches <CR><LF> and 
    {																//	  translates it into a <NL>
    	try {
			StringBuffer buffer = new StringBuffer(256);
            boolean hasDone = false;
			int nextChar;
			
			do  {
            	
            	if( ( nextChar = ccIn.read() ) == '\r' )
                {
                	if( ( nextChar = ccIn.read() ) == '\n' ) {
                		buffer.append( '\n' );
            			if( echo ) terminal.write( "\n" );
                    	hasDone = true;
                    }
                	else {
                		buffer.append( '\r' );
                		buffer.append( (char)nextChar );
            			if( echo ) terminal.write( "\r" + nextChar );
                	}
                }
                else {
                	if( nextChar == -1 ) {
                		//	Se una lettura ritorna la EOF bisogna notificare la caduta della connessione
               			throw new ConnectionException("Control connection interrupted.");
               		}
               		else {
               			buffer.append( (char)nextChar );
            			if( echo ) terminal.write( (char)nextChar );
                	}
                }
            }
            while( !hasDone );
	
			return	buffer.toString();        
        
        } catch( java.io.InterruptedIOException e ) {
            throw new ConnectionException("Control connection interrupted.");
        } catch( java.io.IOException e ) {
            throw new ConnectionException("Unknown control connection exception.");
        } 
    }
        
    ///////////////////////////////////////////////////////////////////////////////
    //
    //	Following are the FTP commands implemented (or wannabe) in this class:
    //  ----------------------------------------------------------------------    
    //
    //		USER <SP> <username> <CRLF>
    //      PASS <SP> <password> <CRLF>
    //      CWD  <SP> <pathname> <CRLF>
    //      CDUP <CRLF>
    //      QUIT <CRLF>
    //      REIN <CRLF>
    //      PORT <SP> <host-port> <CRLF>
    //      PASV <CRLF>
    //      TYPE <SP> <type-code> <CRLF>
    //      STRU <SP> <structure-code> <CRLF>
    //      MODE <SP> <mode-code> <CRLF>
    //      RETR <SP> <pathname> <CRLF>
    //      STOR <SP> <pathname> <CRLF>
    //      STOU <CRLF>
    //      APPE <SP> <pathname> <CRLF>
    //      ALLO <SP> <decimal-integer>
    //          [<SP> R <SP> <decimal-integer>] <CRLF>
    //      REST <SP> <marker> <CRLF>
    //      ABOR <CRLF>
    //      PWD  <CRLF>
    //      LIST [<SP> <pathname>] <CRLF>
    //      SITE <SP> <string> <CRLF>
    //      SYST <CRLF>
    //      STAT [<SP> <pathname>] <CRLF>
    //      HELP [<SP> <string>] <CRLF>
    //      NOOP <CRLF>
    //
    //      FTP COMMAND ARGUMENTS
    //		---------------------
    //      The syntax of the above argument fields 
    //		(using BNF notation where applicable) is:
    //
    //      <username> ::= <string>
    //      <password> ::= <string>
    //      <account-information> ::= <string>
    //      <string> ::= <char> | <char><string>
    //      <char> ::= any of the 128 ASCII characters except <CR> and
    //      <LF>
    //      <marker> ::= <pr-string>
    //      <pr-string> ::= <pr-char> | <pr-char><pr-string>
    //      <pr-char> ::= printable characters, any
    //                    ASCII code 33 through 126
    //      <byte-size> ::= <number>
    //      <host-port> ::= <host-number>,<port-number>
    //      <host-number> ::= <number>,<number>,<number>,<number>
    //      <port-number> ::= <number>,<number>
    //      <number> ::= any decimal integer 1 through 255
    //      <form-code> ::= N | T | C
    //      <type-code> ::= A [<sp> <form-code>]
    //                    | E [<sp> <form-code>]
    //                    | I
    //                    | L <sp> <byte-size>
    //      <structure-code> ::= F | R | P
    //      <mode-code> ::= S | B | C
    //      <pathname> ::= <string>
    //      <decimal-integer> ::= any decimal integer
    // 
	///////////////////////////////////////////////////////////////////////////////
    //
	//		Command-Reply Sequences
	//		-----------------------
	//		In this section, the command-reply sequence is presented.  
	//		Each command is listed with its possible replies; 
	//		command groups are listed together.  Preliminary replies are listed 
	//		first (with their succeeding replies indented and under them), then 
	//		positive and negative completion, and finally intermediary replies 
	//		with the remaining commands from the sequence following.  
	//		This listing forms the basis for the state diagrams, which 
	//		will be presented separately.
	//   
	//		Connection Establishment
	//				120
	//		    		220
	//		    	220
	//		    	421
	//		Login
	//			USER
	//		       	230
	//		       	530
	//		       	500, 501, 421
	//		       	331, 332
	//		    PASS
	//		    	230
	//		       	202
	//		       	530
	//		       	500, 501, 503, 421
	//		       	332
	//		    CWD	
	//		       	250
	//		       	500, 501, 502, 421, 530, 550
	//		    CDUP
	//		       	200
	//		       	500, 501, 502, 421, 530, 550
	//		Logout
	//		    REIN
	//		       	120
	//		        	220
	//		        220
	//		        421
	//		        500, 502
	//		    QUIT
	//		        221
	//		        500
	//		Transfer parameters
	//			PORT
	//		    	200
	//		       	500, 501, 421, 530
	//		    PASV
	//		       	227
	//		       	500, 501, 502, 421, 530
	//		    MODE
	//		       	200
	//		       	500, 501, 504, 421, 530
	//		    TYPE
	//		       	200
	//		       	500, 501, 504, 421, 530
	//		    STRU
	//		       	200
	//		       	500, 501, 504, 421, 530
	//		File action commands
	//		    ALLO
	//		       	200
	//		       	202
	//		       	500, 501, 504, 421, 530
	//		    REST
	//		       	500, 501, 502, 421, 530
	//		       	350
	//		    STOR
	//		       	125, 150
	//		        	(110)
	//		          	226, 250
	//		          	425, 426, 451, 551, 552
	//		    	532, 450, 452, 553
	//		       	500, 501, 421, 530
	//		    STOU
	//		       	125, 150
	//		        	(110)
	//		          	226, 250
	//		          	425, 426, 451, 551, 552
	//		       	532, 450, 452, 553
	//		       	500, 501, 421, 530
	//		    RETR
	//		       	125, 150
	//		          	(110)
	//		          	226, 250
	//		        	425, 426, 451
	//		       	450, 550
	//		       	500, 501, 421, 530
	//		    LIST
	//		       	125, 150
	//		          	226, 250
	//		        	425, 426, 451
	//		       	450
	//		       	500, 501, 502, 421, 530
	//		    NLST
	//		       	125, 150
	//		          	226, 250
	//		          	425, 426, 451
	//		       	450
	//		       	500, 501, 502, 421, 530
	//		    APPE
	//		       	125, 150
	//		          	(110)
	//		          	226, 250
	//		        	425, 426, 451, 551, 552
	//		       	532, 450, 550, 452, 553
	//		       	500, 501, 502, 421, 530
	//		    DELE
	//		       	250
	//		       	450, 550
	//		       	500, 501, 502, 421, 530
	//		    PWD
	//		       	257
	//		       	500, 501, 502, 421, 550
	//		    ABOR
	//		       	225, 226
	//		       	500, 501, 502, 421
	//		    SYST
	//		       	215
	//		       	500, 501, 502, 421
	//		    STAT
	//		       	211, 212, 213
	//		       	450
	//		       	500, 501, 502, 421, 530
	//		    HELP
	//		       	211, 214
	//		       	500, 501, 502, 421
	//		Miscellaneands
	//			SITE
	//		    	200
	//		       	202
	//		       	500, 501, 530
	//		    NOOP
	//		       	200
	//		       	500 421
	//
	///////////////////////////////////////////////////////////////////////////////
	
	private boolean USER( String username ) 
	throws ConnectionException
    {
        String message = "USER " + username;
        WRITELN( message );       
        return REPLY( "331" );
    }
    
    private boolean PASS( String password ) 
    throws ConnectionException
    {
        String message = "PASS " + password;
        WRITELN( message );
        return REPLY( "230" );
    }
    
    private boolean CWD( String pathname ) 
    throws ConnectionException
    {
    	String message = "CWD " + pathname;
    	WRITELN( message );
    	return REPLY( "250" );
    }
    
    private boolean PORT( String addr, int port  ) 
    throws ConnectionException
    {
        int portMSB = ( port >> 8 ) & 0xFF;
        int portLSB =   port        & 0xFF;
        
        String message = "PORT " + addr.replace( '.', ',' ) + ',' + portMSB + ',' + portLSB;
        WRITELN( message );
        return REPLY( "200" );
    }

    private boolean TYPE( String type_code ) 
    throws ConnectionException
    {
        String message = "TYPE " + type_code;
        WRITELN( message );
        return REPLY( "200" );
    }
    
    private boolean STRU( String structure_code ) 
    throws ConnectionException
    {
        String message = "STRU " + structure_code;
        WRITELN( message );
        return REPLY( "200" );
    }

    private boolean MODE( String mode_code ) 
    throws ConnectionException
    {
        String message = "MODE " + mode_code;
        WRITELN( message );
        return REPLY( "200" );
    }

    private boolean RETR( String path_name ) 
    throws ConnectionException
    {
        String message = "RETR " + path_name;
        WRITELN( message );

        //  Expected replies:
        //
        //      125     Data connection already open; transfer starting.
        //      150     File status okay; about to open data connection.
        //     (110)    Restart marker reply.
        //                  In this case, the text is exact and not left to the
        //                  particular implementation; it must read:
        //                  MARK yyyy = mmmm
        //                  Where yyyy is User-process data stream marker, and mmmm
        //                  server's equivalent marker (note the spaces between markers and "=").
        //      226     Closing data connection.
        //                  Requested file action successful (for example, file transfer or file abort).
        //      250     Requested file action okay, completed.
                              
        return REPLY( "150" );
    }

    private boolean REST( long marker )
    throws ConnectionException
    {
        if( marker < 0 )
            throw new ConnectionException("Bad Marker");
            
        String message = "REST " + marker;
        WRITELN( message );
        return REPLY( "350" );
    }
    
    private boolean ABOR()
    throws ConnectionException
    {
        String message = "ABOR";
        WRITELN( message );
        return true;                        // The protocol interpreter is already waiting for the 226 reply
    }
    
    private boolean PWD() 
    throws ConnectionException
    {
        String message = "PWD";
        WRITELN( message );
        return REPLY( "257" );
    }

    private boolean SYST() 
    throws ConnectionException
    {                                       //  returns true if the system of the host is known
        String message = "SYST";			//  detail can be found on the features Properties
        WRITELN( message );					//  under the key "SYST"
        
        if( REPLY("215") ) {    			//  positive reply
            //  TODO:   
            //	read host system information
            return true;
        } else {
        
            return false;
        }
    }

    private boolean NOOP() 
   	throws ConnectionException
	{
        String message = "NOOP";
        WRITELN( message );
        return REPLY( "200" );
    }
    
    ///////////////////////////////////////////////////////////////////////////////
    //
    //		Numeric  Order List of Reply Codes
	//      ----------------------------------
	//
	//      110 Restart marker reply.
	//          In this case, the text is exact and not left to the
	//          particular implementation; it must read:
	//                MARK yyyy = mmmm
	//          Where yyyy is User-process data stream marker, and mmmm
	//          server's equivalent marker (note the spaces between markers
	//          and "=").
	//      120 Service ready in nnn minutes.
	//      125 Data connection already open; transfer starting.
	//      150 File status okay; about to open data connection.
	//      200 Command okay.
	//      202 Command not implemented, superfluous at this site.
	//      211 System status, or system help reply.
	//      212 Directory status.
	//      213 File status.
	//      214 Help message.
	//          On how to use the server or the meaning of a particular
	//          non-standard command.  This reply is useful only to the
	//          human user.
	//      215 NAME system type.
	//          Where NAME is an official system name from the list in the
	//          Assigned Numbers document.
	//      220 Service ready for new user.
	//      221 Service closing control connection.
	//          Logged out if appropriate.
	//      225 Data connection open; no transfer in progress.
	//      226 Closing data connection.
	//          Requested file action successful (for example, file
	//          transfer or file abort).
	//      227 Entering Passive Mode (h1,h2,h3,h4,p1,p2).
	//      230 User logged in, proceed.
	//      250 Requested file action okay, completed.
	//      257 "PATHNAME" created.
	//      331 User name okay, need password.
	//      332 Need account for login.
	//      350 Requested file action pending further information.
	//      421 Service not available, closing control connection.
	//          This may be a reply to any command if the service knows it
	//          must shut down.
	//      425 Can't open data connection.
	//      426 Connection closed; transfer aborted.
	//      450 Requested file action not taken.
	//          File unavailable (e.g., file busy).
	//      451 Requested action aborted: local error in processing.
	//      452 Requested action not taken.
	//          Insufficient storage space in system.
	//      500 Syntax error, command unrecognized.
	//          This may include errors such as command line too long.
	//      501 Syntax error in parameters or arguments.
	//      502 Command not implemented.
	//      503 Bad sequence of commands.
	//      504 Command not implemented for that parameter.
	//      530 Not logged in.
	//      532 Need account for storing files.
	//      550 Requested action not taken.
	//          File unavailable (e.g., file not found, no access).
	//      551 Requested action aborted: page type unknown.
	//      552 Requested file action aborted.
	//          Exceeded storage allocation (for current directory or
	//          dataset).
	//      553 Requested action not taken.
	//          File name not allowed.
  	//
	////////////////////////////////////////////////////////////////////////////////
	
  	private boolean REPLY( String expected )
	throws ConnectionException
    {
        String  reply;
        String  replyCode;
        boolean replyValue;
        boolean hasDone;

        if( (reply = READLN() ) != null ) 
        {
            if( (replyCode = reply.substring(0,4)).startsWith(expected) ) {          
                replyValue = true;
            } else {
                replyValue = false;
            }
    
            lastReply = new Vector();				//	Qui si puo ottimizare un po...
            lastReply.addElement( reply );
        
            hasDone = replyCode.endsWith(" ");

            while(!hasDone) {
                if( (reply = READLN() ) != null ) {
                    lastReply.addElement( reply );
                    hasDone = (replyCode = reply.substring(0,4)).endsWith(" ");
                } else {
                    //  TODO:   handle unexpected eof
                    return  false;
                }
            }
        
            return  replyValue;
        
        } else {
            // TODO:    handle unexpected eof
            return  false;
        }
    }
    
    private boolean FEAT() 
    throws ConnectionException
    {
        String message = "FEAT";
        WRITELN( message );
        return REPLY( "211" );    
    }
    
    private boolean SIZE( String path ) 
    throws ConnectionException
    {
        String message = "SIZE " + path;
        WRITELN( message );
        return REPLY( "213" );    
    }
    
    private boolean MDTM( String path ) 
    throws ConnectionException
    {
        String message = "MDTM " + path;
        WRITELN( message );
        return REPLY( "213" );    
    }
    
		//{{DECLARE_CONTROLS
		//}}
}    